/*
 * cntlr.h - MAP controller header file
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 */

#ifndef CNTLR_H
#define CNTLR_H

#define cntlr_dbg(...)     dbg(green "CNTLR: " nocl __VA_ARGS__)
#define cntlr_warn(...)    warn(red "CNTLR: " __VA_ARGS__)
#define cntlr_info(...)    info(blue "CNTLR: " nocl __VA_ARGS__)

#include <timer_impl.h>
#include <cmdu_ackq.h>
#include <map_module.h>

#include <json-c/json.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>
#include <libubus.h>

#include "config.h"
#include "steer.h"

#ifndef EASYMESH_VENDOR_EXT_OUI
#define EASYMESH_VENDOR_EXT_OUI         (uint8_t *)"\x00\x11\x22"
#endif

extern const char *ubus_socket;

typedef uint32_t object_t;
#define OBJECT_INVALID	((uint32_t)-1)

enum iface_type {
	NETIF_FHBSS,
	NETIF_BKBSS,
	NETIF_BSTA
};

enum device_type {
	NON_IEEE1905,
	IEEE1905
};

struct cac_data {
	uint8_t radio[6];
	uint8_t opclass;
	uint8_t channel;
	uint8_t cac_method;
	uint8_t cac_action;
};

struct bcn_metrics {
	uint8_t opclass;
	uint8_t channel;
	uint8_t rcpi;
	uint8_t rsni;
	uint8_t bssid[6];
	bool requested; /* beacon metrics query sent by cntlr */
	bool stale;	/* outdated report got before most recent request sent */
	struct list_head list;
};

struct bcn_meas_element {
	uint8_t tag_number;
	uint8_t tag_length;
	uint8_t meas_token;
	uint8_t meas_report_mode;
	uint8_t meas_report_type;
	uint8_t op_class;
	uint8_t channel;
	uint8_t start_time[8];
	uint8_t duration[2];
	uint8_t frame_info;
	uint8_t rcpi;
	uint8_t rsni;
	uint8_t bssid[6];
	uint8_t antena_id;
	uint8_t tsf[4];
	uint8_t frame[];
};

struct bcnreq {
	uint8_t sta_mac[6];
	uint8_t agent_mac[6];
	struct timespec tsp;
	int request_num;

	struct list_head list;
};

struct una_sta_metrics {
	/* node the measurement has been done on */
	struct node *agent;
	uint8_t channel;
	uint32_t time_delta;
	uint8_t ul_rcpi;

	struct list_head list;
};

enum steer_trigger {
	STEER_TRIGGER_UNKNOWN,
	STEER_TRIGGER_UTIL,
	STEER_TRIGGER_LINK_QUALITY,
	STEER_TRIGGER_BK_UTIL
};

enum steer_method {
	STEER_METHOD_UNKNOWN,
	STEER_METHOD_ASYNC_BTM,
	STEER_METHOD_BTM_REQ,
	STEER_METHOD_ASSOC_CTL
};

struct steer_attempt {
	struct timespec time;
	uint8_t src_bssid[6];
	uint8_t dst_bssid[6];
	enum steer_trigger trigger;
	enum steer_method method;
	uint32_t duration; /* seconds */
	bool auto_bssid;
	bool answered;
};

#define STEER_STATS_NO_DATA 0xffffffffffffffff
struct steer_stats {
	/* The maximum value that can be represented as an unsignedLong
	 * (i.e. 0xffffffffffffffff) indicates that no data is available
	 * for this parameter.
	 */
	uint64_t no_candidate_cnt;
	uint64_t blacklist_attempt_cnt;
	uint64_t blacklist_success_cnt;
	uint64_t blacklist_failure_cnt;
	uint64_t btm_attempt_cnt;
	uint64_t btm_success_cnt;
	uint64_t btm_failure_cnt;
	uint64_t btm_query_resp_cnt;
	/* Reperesents time this Associated Device was last attempted
	 * to be steered. Calculate delta from now (secs) for reporting
	 */
	struct timespec last_attempt_tsp;
};

struct sta {
	uint8_t macaddr[6];
	uint8_t bssid[6];
	uint16_t conntime;

	uint32_t time_delta;
	uint32_t ul_thput;
	uint32_t dl_thput;
	uint8_t ul_rcpi;

	uint8_t num_bcn_metrics;
	//struct bcn_metrics bcns[8];

	struct list_head bcnlist;

	struct list_head unassoclist;

	struct netif_iface *fh; /* the AP interface the sta is connected at */

	enum device_type type;
	struct node *agent; /* the agent representing an IEEE1905 device, only applicable for backhaul stations */
	struct uloop_timeout bcn_metrics_timer;
	struct timespec last_bcn_metrics_query;
#define BCN_REPORT_WAIT_MAX 60
#define BCN_REPORT_STALE_TIME 20
	int bcn_report_wait_time; /* maximum time to wait for bcn response in seconds */

	int failed_steer_attempts;
	int pending_btm_resp_num;

	uint8_t num_steer_attempts;
#define MAX_LOG_STEER_ATTEMPTS 10
	struct steer_attempt attempts[MAX_LOG_STEER_ATTEMPTS];

	struct steer_stats stats;

#define BTM_RESP_EXP_TIMEOUT 5
	atimer_t btm_req_timer;

	struct list_head list;
};

/** Latest combined infra metrics data **/
struct bss_metrics {
	// struct node *n;
	// struct list_head list;
	uint8_t bssid[6];
	// bool active;	/* for add to list */
	uint8_t ch_util;
	uint16_t num_sta;	/** Number of STAs currently associated */
#define ESP_AC_BE	0x80
#define ESP_AC_BK	0x40
#define ESP_AC_VO	0x20
#define ESP_AC_VI	0x10
	uint8_t esp_ac;	/** Estimated Service Parameters Information */
	uint8_t esp_be[3];
	uint8_t esp[];
	// uint8_t upstream_bssid[6];
};

enum media_type {
	ETH,
	WIFI
};

/* This struct maps to fronthaul AP-BSS interface
 * of neighboring nodes.
 */
struct netif_iface {
	char ifname[16];
	char ssid[33];
	int band;
	enum media_type med_type;
	uint32_t bssid_info;
	uint8_t bssid[6];
	uint8_t reg;
	uint8_t phy;
	int channel;
	int capacity;
	int type;
	bool active;
	struct timespec tsp;
	struct node *agent;
	struct bss_metrics *metrics;	/** AP Metrics TLV data **/
	struct list_head list;
	uint8_t upstream_bssid[6]; /* in the case of the interface is a bsta interface */
};

/** Latest combined infra metrics data **/
struct link_metrics {
	// unsigned char downstream[6];
	// unsigned char upstream[6];
#define NETIF_LINK_ETH	1
#define NETIF_LINK_WIFI	2
	uint16_t type;	/** Media Type*/
	/* tx link metrics */
	bool bridge;
	uint32_t packet_tx_error;
	uint32_t packet_trans;
	uint16_t thp;	/** MAC Throughput Capacity in Mbps */
	uint16_t link_av;	/** Link Availability */
	uint16_t phy_rate;	/* in Mpbs */
	/* rx link metrics */
	uint32_t packet_rx_error;
	uint32_t packet_rec;
	int8_t rssi;	/** in dBm */
	struct netif_link *l;
	struct list_head list;
};

/* cut-down version of what agents have */
/* TODO - refine this struct */
struct netif_link {
//	char name[16];
	struct netif_iface *downstream;
	struct netif_iface *upstream;
	int channel;
#define NETIF_LINK_ETH	1
#define NETIF_LINK_WIFI	2
	/* int type; */
	/* bool active; */
	int capacity;
	struct link_metrics *metrics;	/** 1905 Link Metric TLV data **/
	struct list_head list;
};

struct opclass_entry {
	uint8_t opclass;
	int bw;
	int txpower;

	int channels_num;
	struct {
		uint8_t channel;
		uint8_t preference;
	} channels[64];
};

struct opclass {
	struct opclass_entry opclass_entry[64];
	int opclass_entry_num;
	time_t opclass_entry_time;
};

/* TODO - fill this structure */
struct netif_radio {
	char name[16];
	uint8_t macaddr[6];
	enum wifi_freqband band;
	bool isup;

	struct opclass cur_opclass;
	struct opclass opclass;

	struct node *agent;
	struct list_head iflist; /** list of netif_fhbss */
	struct list_head list;
};

enum nodetype {
	NODE_WIFI_EXTENDER,
	NODE_WIFI_REPEATER,
	NODE_ETH_EXTENDER,
};


enum uplink_type {
	UL_UNSPEC,
	UL_ETH,
	UL_WIFI,
	UL_DSL,
};

struct uobj_struct {
	object_t id;
	int (*req)(int argc, char **argv);
	void (*resp)(struct ubus_request *r, int t, struct blob_attr *m);
};

struct watchnode {
	unsigned char hwaddr[6];
	struct in_addr ipaddr;
	/* char fh_ifname[16]; */
	unsigned char fh_bssid[6];
	char fh_ssid[33];
	struct timeval tv;
#define WATCHNODE_SCAN_RETRY_MAX   5
	unsigned int scan_retry;
	struct uloop_timeout scan_timer;
	struct uloop_timeout scanres_timer;
	struct controller *cntlr;
	struct list_head list;
};

/* struct node - maps to a 1905 device */
struct node {
	uint8_t alid[6];
	struct in_addr ipaddr;
	enum nodetype type;
	int depth;                         /** >= 0 or -1 for unknown */
	int ul_type;                       /** uplink type */
	bool scan_supported;               /** whether scanning supported */
	struct controller *cntlr;
#define MAX_UOBJECTS	8
	struct uloop_timeout refresh_timer;
	struct ubus_event_handler evh;
	//struct agent_policy *ap;
	struct node_policy *np;
	uint8_t ap_cap;
	struct list_head radiolist;        /** list of netif_radio */
	struct list_head stalist;
	struct list_head list;
};

#define NODE_STATS_INTERVAL		30000       /** 30 secs */

enum cntlr_state {
	CNTLR_INIT, /* while initializing, only look for ACS answer */
	CNTLR_IDLE, /* TODO: while idle, do nothing in a loop till awakened */
	CNTLR_START /* normal */
};

struct steering {
	int channels_num;
#define STA_CHANNEL_REPORT_MAX_CHAN 64
	uint8_t channels[STA_CHANNEL_REPORT_MAX_CHAN];
};

struct controller {
	enum cntlr_state state;
	unsigned char almac[6];
	void *comm;
	struct ubus_object obj;
	struct ubus_context *ubus_ctx;
	struct ubus_event_handler evh;
	struct uloop_timeout heartbeat;
	struct uloop_timeout acs;
	struct uloop_timeout dfs_cleanup;
	int num_nodes;
	int num_tx_links;
	int num_rx_links;
	struct list_head nodelist;
	struct list_head stalist;
	struct list_head bcnreqlist;
	struct list_head linklist;
	struct uloop_timeout radar_timer;
	struct uloop_timeout discovery_timer;
	struct uloop_timeout start_timer;
	struct uloop_timeout signal_handler;
	struct uloop_timeout query_nodes;
	struct hlist_head *as_table; /** active sta hash table */
	struct controller_config cfg;
	struct cmdu_ackq cmdu_ack_q;

	/* Autoconfig */
	uint16_t mid_5g;
	uint16_t mid_2g;

	/* i1905.map registration */
	uint32_t map_oid;
	mapmodule_cmdu_mask_t cmdu_mask;
	void *subscriber;
	bool subscribed;

	struct list_head sclist;	/* steer-control module list */
	struct steer_control *sctrl;	/* active steer-control module */
};

struct sta_channel_report {
	uint8_t opclass;
	uint8_t num_channel;
	uint8_t channel[128];
};

#define MAX_UNASSOC_STAMACS 10
struct unassoc_sta_metric {
	uint8_t channel;
	uint8_t num_sta;
	struct {
		uint8_t macaddr[6];
	} sta[MAX_UNASSOC_STAMACS];
};

struct sta_error_response {
	uint8_t sta_mac[6];
	uint8_t response;
};

#define COMM_HANDLE(c)	(((struct controller *)(c))->ubus_ctx)

struct node *alloc_node_init(struct controller *c, uint8_t *hwaddr);
struct netif_iface *find_interface_by_ssid(struct controller *c,
		struct node *n, char *ssid);
struct netif_radio *find_radio_by_ssid(struct controller *c,
		struct node *n, char *ssid);
struct netif_link *alloc_link_init(struct controller *c,
		uint8_t *upstream, uint8_t *downstream);
struct netif_radio *find_radio_by_node(struct controller *c, struct node *n,
		uint8_t *radio);
struct netif_radio *find_radio_by_bssid(struct controller *c, uint8_t *bssid);
struct node *find_node_by_mac(struct controller *c, uint8_t *mac);
struct node *find_node_by_iface(struct controller *c, uint8_t *bssid);
struct netif_iface *find_interface_by_mac(struct controller *c,
		struct netif_radio *r, uint8_t *hwaddr);
struct netif_iface *find_interface_by_mac_nor(struct controller *c,
		uint8_t *hwaddr);
struct node_policy *agent_find_policy(struct controller *c, uint8_t *agent);
struct radio_policy *agent_find_radio_policy(struct controller *c, uint8_t *bssid);
struct netif_radio *cntlr_node_add_radio(struct controller *c, struct node *n,
		uint8_t *radio);
struct netif_iface *cntlr_radio_add_interface(struct controller *c,
		struct netif_radio *r, uint8_t *hwaddr);
struct netif_iface *cntlr_iterate_fbss(struct controller *c, uint8_t *mac);
void cntlr_update_sta_steer_counters(struct controller *c, uint8_t *sta_mac,
		uint8_t *src_bssid, uint8_t *dst_bssid,
		uint32_t mode, enum steer_trigger trigger);
void cntlr_notify_client_steer_req_evt(struct controller *c,
		uint8_t *bssid, uint32_t sta_nr, uint8_t stas[][6],
		uint32_t bssid_nr, uint8_t target_bssid[][6]);
struct sta *cntlr_add_sta(struct controller *c, uint8_t *macaddr);
struct sta *cntlr_find_sta(struct controller *c, uint8_t *mac);
struct bcnreq *cntlr_find_bcnreq(struct controller *c, uint8_t *sta, uint8_t *alid);
struct netif_iface *cntlr_get_fbss_by_mac(struct controller *c, struct node *n,
		uint8_t *mac);
bool cntlr_resync_config(struct controller *c, bool reload);
void free_bcn_metrics(struct controller *c, struct sta *s);
void free_usta_metrics(struct controller *c, struct sta *s);

int cntlr_radio_opclass_add(struct netif_radio *radio, uint8_t opclass,
			    uint8_t channel, uint8_t preference);
void cntlr_radio_opclass_reset(struct netif_radio *radio);
void cntlr_radio_opclass_dump(struct netif_radio *radio);

int cntlr_radio_cur_opclass_add(struct netif_radio *radio, uint8_t opclass,
				uint8_t channel, uint8_t txpower);
void cntlr_radio_cur_opclass_reset(struct netif_radio *radio);
void cntlr_radio_cur_opclass_dump(struct netif_radio *radio);
void cntlr_radio_update_band(struct netif_radio *radio);
bool cntlr_node_opclass_expired(struct node *node);

int cntlr_sync_dyn_controller_config(struct controller *c, uint8_t *agent);

void cntlr_load_steer_modules(struct controller *c);
void cntlr_unload_steer_modules(struct controller *c);

#endif /* CNTLR_H */
